import os
import unittest

from ..util import NO_DEFAULT, get_file_name


class TestUtil(unittest.TestCase):

    def test_NO_DEFAULT_is_False(self):
        self.assertFalse(NO_DEFAULT)

    def test_default_file_name(self):
        os.environ.pop('LOCAL_SETTINGS_FILE', None)
        file_name = get_file_name()
        self.assertEqual(os.path.basename(file_name), 'local.cfg')

    def test_set_file_name_via_environ(self):
        os.environ['LOCAL_SETTINGS_FILE'] = 'pants.cfg'
        file_name = get_file_name()
        self.assertEqual(file_name, 'pants.cfg')



import json
import os
import pkg_resources
import sys

from .color_printer import color_printer as printer
from .exc import SettingsFileDidNotPassCheck
from .loader import Loader
from .strategy import INIJSONStrategy
from .util import get_file_name

# Exported (but unused locally)
from .checker import Checker  # noqa: exported
from .exc import SettingsFileNotFoundError  # noqa: exported
from .settings import Settings  # noqa: exported
from .types import LocalSetting, SecretSetting  # noqa: exported
from .util import NO_DEFAULT  # noqa: exported
from .__main__ import make_local_settings  # noqa: exported


def load_and_check_settings(base_settings, file_name=None, section=None, base_path=None,
                            strategy_type=INIJSONStrategy, disable=None, prompt=None,
                            quiet=None):
    """Merge local settings from file with base settings, then check.

    Returns a new dict containing the base settings and the loaded
    settings. Includes:

        - base settings
        - settings from extended file(s), if any
        - settings from file

    Settings loaded from the specified file will override base settings,
    then the settings will be checked to ensure that all required local
    settings have been set.

    If a file name is passed: if the file exists, local settings will be
    loaded from it and any missing settings will be appended to it; if
    the file does not exist, it will be created and all settings will be
    added to it.

    If a file name isn't passed: if the ``LOCAL_SETTINGS_FILE_NAME``
    environment variable is set, the specified file will be used;
    otherwise ``{base_path}/local.cfg`` will be used.

    ``base_path`` is used when ``file_name`` is relative; if it's not
    passed, it will be set to the current working directory.

    When ``prompt`` is ``True``, the user will be prompted for missing
    local settings. By default, the user is prompted only when running
    on TTY. The ``LOCAL_SETTINGS_CONFIG_PROMPT`` environment variable
    can be used to set ``prompt``.

    When ``quiet`` is ``True``, informational messages will not be
    printed. The ``LOCAL_SETTINGS_CONFIG_QUIET`` environment variable
    can be used to set ``quiet``.

    .. note:: When setting flags via environment variables, use a JSON
        value like 'true', '1', 'false', or '0'.

    See :meth:`.Loader.load` and :meth:`.Checker.check` for more info.

    """
    environ_config = get_config_from_environ()
    disable = environ_config['disable'] if disable is None else disable
    prompt = environ_config['prompt'] if prompt is None else prompt
    quiet = environ_config['quiet'] if quiet is None else quiet
    if disable:
        return {}
    if file_name is None:
        file_name = get_file_name()
    if ':' in file_name:
        package, path = file_name.split(':', 1)
        file_name = pkg_resources.resource_filename(package, path)
    if not os.path.isabs(file_name):
        base_path = base_path or os.getcwd()
        file_name = os.path.normpath(os.path.join(base_path, file_name))
    try:
        loader = Loader(file_name, section, strategy_type=strategy_type)
        settings, success = loader.load_and_check(base_settings, prompt)
    except KeyboardInterrupt:
        # Loading/checking of local settings was aborted with Ctrl-C.
        # This isn't an error, but we don't want to continue.
        if not quiet:
            printer.print_warning('\nAborted loading/checking of local settings')
        sys.exit(0)
    if loader.section:
        file_name = '{loader.file_name}#{loader.section}'.format(loader=loader)
    else:
        file_name = loader.file_name
    if not success:
        raise SettingsFileDidNotPassCheck(file_name)
    if not quiet:
        printer.print_success('Settings loaded successfully from {0}'.format(file_name))
    return settings


def get_config_from_environ():
    def get(name, default='null'):
        name = name.upper()
        name = 'LOCAL_SETTINGS_CONFIG_{name}'.format(name=name)
        return json.loads(os.environ.get(name, default))
    options = (
        ('disable', 'false'),
        ('prompt', 'null'),
        ('quiet', 'false'),
    )
    return {n: get(n, default) for (n, default) in options}



import unittest

from ..exc import NoDefaultError
from ..types import LocalSetting


class TestLocalSetting(unittest.TestCase):

    def test_create_local_setting(self):
        setting = LocalSetting()
        self.assertRaises(NoDefaultError, lambda: setting.default)
        self.assertFalse(setting.has_default)
        self.assertFalse(setting.has_value)

    def test_create_local_setting_with_default(self):
        setting = LocalSetting(default={})
        self.assertTrue(setting.has_default)
        self.assertEqual(setting.default, {})
        self.assertTrue(setting.has_value)
        self.assertEqual(setting.value, {})

    def test_set_value_of_local_setting_with_default(self):
        setting = LocalSetting(default={})
        setting.value = 'X'
        self.assertEqual(setting.default, {})
        self.assertEqual(setting.value, 'X')

    def test_local_setting_default_must_be_json_serializable(self):
        self.assertRaises(TypeError, LocalSetting, default=object())

    def test_local_setting_can_has_local_setting_as_default(self):
        default_setting = LocalSetting(default=1)
        setting = LocalSetting(default=default_setting)
        self.assertTrue(setting.has_default)
        self.assertEqual(setting.default, 1)
        self.assertEqual(setting.value, 1)
        setting.value = 2
        self.assertEqual(default_setting.value, 1)
        self.assertEqual(setting.value, 2)

    def test_local_setting_with_validator(self):
        setting = LocalSetting(validator=lambda v: isinstance(v, int))
        with self.assertRaises(ValueError):
            setting.value = 'abc'
        setting.value = 123

    def test_local_setting_with_callable_default(self):
        setting = LocalSetting(default=lambda: 'pants')
        self.assertEqual(setting.default, 'pants')



"""Strategies for reading from & writing to config files."""
import logging
import json
import os
import pkg_resources
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from configparser import NoSectionError, RawConfigParser

from six import raise_from, with_metaclass

from .exc import SettingsFileNotFoundError, SettingsFileSectionNotFoundError


__all__ = [
    'Strategy',
    'INIStrategy',
    'INIJSONStrategy',
]


log = logging.getLogger(__name__)


class Strategy(with_metaclass(ABCMeta)):

    file_types = ()

    @abstractmethod
    def read_file(self, file_name, section=None):
        """Read settings from file."""

    @abstractmethod
    def write_settings(self, settings, file_name, section=None):
        """Write settings to file."""

    def parse_file_name_and_section(self, file_name, section=None, extender=None,
                                    extender_section=None):
        """Parse file name and (maybe) section.

        File names can be absolute paths, relative paths, or asset
        specs::

            /home/user/project/local.cfg
            local.cfg
            some.package:local.cfg

        File names can also include a section::

            some.package:local.cfg#dev

        If a ``section`` is passed, it will take precedence over a
        section parsed out of the file name.

        """
        if '#' in file_name:
            file_name, parsed_section = file_name.rsplit('#', 1)
        else:
            parsed_section = None

        if ':' in file_name:
            package, path = file_name.split(':', 1)
            file_name = pkg_resources.resource_filename(package, path)

        if extender:
            if not file_name:
                # Extended another section in the same file
                file_name = extender
            elif not os.path.isabs(file_name):
                # Extended by another file in the same directory
                file_name = os.path.join(os.path.dirname(extender), file_name)

        if section:
            pass
        elif parsed_section:
            section = parsed_section
        elif extender_section:
            section = extender_section
        else:
            section = self.get_default_section(file_name)

        return file_name, section

    def get_default_section(self, file_name):
        return None

    def decode_value(self, value):
        """Decode value read from file to Python object."""
        return value

    def encode_value(self, value):
        """Encode Python object as value that can be written to file."""
        return value


class LocalSettingsConfigParser(RawConfigParser):

    def options(self, section):
        # Order [DEFAULT] options before section options; the default
        # implementation orders them after.
        options = self._defaults.copy()
        try:
            options.update(self._sections[section])
        except KeyError:
            raise_from(NoSectionError(section), None)
        return list(options.keys())

    def optionxform(self, option):
        # Don't alter option names; the default implementation lower
        # cases them.
        return option


class INIStrategy(Strategy):

    file_types = ('ini',)

    def __init__(self):
        super(INIStrategy, self).__init__()
        self.section_found_while_reading = False

    def read_file(self, file_name, section=None):
        """Read settings from specified ``section`` of config file."""
        file_name, section = self.parse_file_name_and_section(file_name, section)
        if not os.path.isfile(file_name):
            raise SettingsFileNotFoundError(file_name)
        parser = self.make_parser()
        with open(file_name) as fp:
            parser.read_file(fp)

        settings = OrderedDict()

        if parser.has_section(section):
            section_dict = parser[section]
            self.section_found_while_reading = True
        else:
            section_dict = parser.defaults().copy()

        extends = section_dict.get('extends')

        if extends:
            extends = self.decode_value(extends)
            extends, extends_section = self.parse_file_name_and_section(
                extends, extender=file_name, extender_section=section)
            settings.update(self.read_file(extends, extends_section))

        settings.update(section_dict)

        if not self.section_found_while_reading:
            raise SettingsFileSectionNotFoundError(section)

        return settings

    def write_settings(self, settings, file_name, section):
        file_name, section = self.parse_file_name_and_section(file_name, section)
        parser = self.make_parser()
        if os.path.exists(file_name):
            with open(file_name) as fp:
                parser.read_file(fp)
        else:
            log.info('Creating new local settings file: %s', file_name)
        if section not in parser:
            log.info('Adding new section to %s: %s', file_name, section)
            parser.add_section(section)
        sorted_keys = sorted(settings.keys())
        for name in sorted_keys:
            value = self.encode_value(settings[name])
            settings[name] = value
            parser[section][name] = value
        with open(file_name, 'w') as fp:
            parser.write(fp)
        for name in sorted_keys:
            value = settings[name]
            log.info('Saved %s to %s as: %s', name, file_name, value)

    def get_default_section(self, file_name):
        """Returns first non-DEFAULT section; falls back to DEFAULT."""
        if not os.path.isfile(file_name):
            return 'DEFAULT'
        parser = self.make_parser()
        with open(file_name) as fp:
            parser.read_file(fp)
        sections = parser.sections()
        section = sections[0] if len(sections) > 0 else 'DEFAULT'
        return section

    def make_parser(self, *args, **kwargs):
        return LocalSettingsConfigParser(*args, **kwargs)


class INIJSONStrategy(INIStrategy):

    file_types = ('cfg',)

    def decode_value(self, value):
        value = value.strip()
        if not value:
            return ''
        try:
            value = json.loads(value)
        except ValueError:
            raise ValueError('Could not parse `{value}` as JSON'.format(**locals()))
        return value

    def encode_value(self, value):
        return json.dumps(value)


def get_strategy_types():
    """Get a list of all :class:`Strategy` subclasses."""
    def get_subtypes(type_):
        subtypes = type_.__subclasses__()
        for subtype in subtypes:
            subtypes.extend(get_subtypes(subtype))
        return subtypes
    return get_subtypes(Strategy)


def get_file_type_map():
    """Map file types (extensions) to strategy types."""
    file_type_map = {}
    for strategy_type in get_strategy_types():
        for ext in strategy_type.file_types:
            if ext in file_type_map:
                raise KeyError(
                    'File type {ext} already registered to {file_type_map[ext]}'
                    .format(**locals()))
            file_type_map[ext] = strategy_type
    return file_type_map


def guess_strategy_type(file_name_or_ext):
    """Guess strategy type to use for file by extension.

    Args:
        file_name_or_ext: Either a file name with an extension or just
            an extension

    Returns:
        Strategy: Type corresponding to extension or None if there's no
            corresponding strategy type

    """
    if '.' not in file_name_or_ext:
        ext = file_name_or_ext
    else:
        name, ext = os.path.splitext(file_name_or_ext)
    ext = ext.lstrip('.')
    file_type_map = get_file_type_map()
    return file_type_map.get(ext, None)



import os
import unittest

from ..loader import Loader
from ..types import LocalSetting


LOCAL_SETTINGS_FILE = os.path.join(os.path.dirname(__file__), 'local.cfg#test')
DERIVED_LOCAL_SETTINGS_FILE = os.path.join(os.path.dirname(__file__), 'local.derived.cfg')


class TestLoading(unittest.TestCase):

    def setUp(self):
        self.loader = Loader(LOCAL_SETTINGS_FILE)

    def test_loading(self):
        local_setting = LocalSetting('default value')
        self.assertEqual(local_setting.default, 'default value')
        self.assertEqual(local_setting.value, 'default value')

        settings = self.loader.load({
            'BASE_SETTING': '{{PACKAGE}}',
            'LOCAL_SETTING': local_setting,
            'TUPLE_SETTING': ('a', 'b', 'c'),
            'LIST_SETTING': ['a', 'b', 'c'],
            'DICT_SETTING': {'a': 'b'},
            'NESTED_SETTING': {
                'list': ['a', 'b'],
                'tuple': ('a', 'b'),
                'dict': {
                    'list': [1, 2],
                    'tuple': (1, 2),
                }
            }
        })

        expected = {
            'BASE_SETTING': 'local_settings',
            'LOCAL_SETTING': 'local value',
            'TUPLE_SETTING': ('a', 'b', 'c'),
            'LIST_SETTING': ['a', 'b', 'c'],
            'DICT_SETTING': {'a': 'b'},
            'NESTED_SETTING': {
                'list': ['a', 'b'],
                'tuple': ('a', 'b'),
                'dict': {
                    'list': [1, 2],
                    'tuple': (1, 2),
                }
            },

            'PACKAGE': 'local_settings',

            'A': {
                'b': {
                    'c': 1,
                    'd': 2,
                },
            },

            'X': {
                'y': {
                    'z': 'z',
                },
            },

            'LIST': ['a', 'b'],
            'OTHER_LIST': ['a', 'b'],

            'TEMPLATES': [
                {
                    'BACKEND': 'package.module.Class',
                    'OPTIONS': {
                        'context_processors': ['a.b', 'x.y.z'],
                    },
                },
            ],

            'INTERPOLATED': {
                'x': 'value',
                'y': 'value',
                'z': 'value',
            },

            'value': 'interpolated key',
            'I': {
                'value': 'interpolated key',
            },
            'Z': [{'value': 1}],
            'J': {
                'local_settings': {
                    'K': 1,
                },
                'local_settingsXXX': {
                    'L': 2,
                },
            },

            'STUFF': ['a prepended thing', 'thing', 'another thing'],

            'DEFAULT_ITEMS': ['0', 'b', 'c'],

            'BASE': {
                'setting': 1,
                'another_setting': 2,
            },

            'OS': {
                'PATH': os.path,
            },

            'FORMAT_STRING': '1{format}',

            'NUMBER': 1,
            'OTHER_NUMBER': 1,

            'LIST1': [1, 2],
            'LIST2': [1, 2],
            'LIST3': [1, 2],

            'DICT': {'1': 1, '2': [[1, 2]]},
            'OTHER_DICT': {'1': 1, '2': [[1, 2]]},
        }
        self.assertEqual(local_setting.default, 'default value')
        self.assertEqual(local_setting.value, 'local value')
        self.assertEqual(settings.INTERPOLATED.x, 'value')
        self.assertEqual(settings.get_dotted('INTERPOLATED.x'), 'value')
        self.assertEqual(settings.DEFAULT_ITEMS, expected['DEFAULT_ITEMS'])
        self.assertEqual(settings, expected)


class TestLoadTypes(unittest.TestCase):

    def setUp(self):
        settings_file = os.path.join(os.path.dirname(__file__), 'local.cfg#test:empty')
        self.loader = Loader(settings_file)

    def test_load_list(self):
        settings = self.loader.load({
            'LIST': ['a', 'b'],
        })
        self.assertIsInstance(settings.LIST, list)

    def test_load_2_tuple(self):
        settings = self.loader.load({
            'TUPLE': ('a', 'b'),
        })
        self.assertIsInstance(settings.TUPLE, tuple)
        self.assertEqual(settings.TUPLE, ('a', 'b'))

    def test_load_3_tuple(self):
        settings = self.loader.load({
            'TUPLE': ('a', 'b', 'c'),
        })
        self.assertIsInstance(settings.TUPLE, tuple)
        self.assertEqual(settings.TUPLE, ('a', 'b', 'c'))

    def test_load_nested_tuple(self):
        settings = self.loader.load({
            'DICT': {
                'TUPLE': ('a', 'b', 'c'),
            }
        })
        self.assertIsInstance(settings.DICT.TUPLE, tuple)
        self.assertEqual(settings.DICT.TUPLE, ('a', 'b', 'c'))

    def test_load_other_sequence_type(self):
        MyList = type('MyList', (list,), {})
        settings = self.loader.load({
            'MY_LIST': MyList(('a', 'b', 'c')),
        })
        self.assertIsInstance(settings.MY_LIST, MyList)
        self.assertEqual(settings.MY_LIST, ['a', 'b', 'c'])


class TestLoadingDerivedSettings(unittest.TestCase):

    def test_loading_derived_settings(self):
        loader = Loader(DERIVED_LOCAL_SETTINGS_FILE)
        settings = loader.load({})
        self.assertEqual(settings.BASE.setting, 1)  # not overridden
        self.assertEqual(settings.BASE.another_setting, "overridden")  # not overridden
        self.assertEqual(settings.DEFAULT_ITEMS, ['first', 'b', 'c'])

    def test_loading_derived_settings_where_section_is_not_present_in_derived_settings_file(self):
        loader = Loader(DERIVED_LOCAL_SETTINGS_FILE + '#test:1')
        settings = loader.load({})
        self.assertEqual(settings.ITEM, "item")



from __future__ import print_function

import sys

from six import add_metaclass

from .util import is_a_tty


class ColorPrinterMeta(type):

    def __init__(cls, *args, **kwargs):
        def _make_methods(_color):
            def _print(self, *a, **kw):
                kw['color'] = _color
                return self.print(*a, **kw)

            def _string(self, *a, **kw):
                return self.string(_color, *a, **kw)

            setattr(cls, 'print_{0}'.format(color), _print)
            setattr(cls, 'string_{0}'.format(color), _string)

        for color in cls.colors:
            _make_methods(color)


@add_metaclass(ColorPrinterMeta)
class ColorPrinter(object):

    """Prints things in color (or not).

    Default colors can be overridden by passing a dict of colors to
    the constructor.

    Use stand-alone or as a mixin::

        >>> printer = ColorPrinter()
        >>> printer.print('boring old message')
        boring old message
        >>> printer.string('none', 'boring old message')
        'boring old message\\x1b[0m'
        >>> printer.print_info('check this out')
        check this out
        >>> printer.print_error('whoopsie')
        whoopsie
        >>> printer.string('error', 'whoopsie')
        '\\x1b[91mwhoopsie\\x1b[0m'
        >>> MyClass = type('MyClass', (ColorPrinter,), {})
        >>> my_obj = MyClass(colors={'header': '\033[96m'})
        >>> my_obj.print_header('Header')
        Header
        >>> my_obj.string('header', 'Header')
        '\\x1b[96mHeader\\x1b[0m'

    Note: This uses the print function from Python 3.

    """

    colors = {
        'header': '\033[95m',
        'info': '\033[94m',
        'success': '\033[92m',
        'warning': '\033[93m',
        'error': '\033[91m',
        'reset': '\033[0m',
        'none': '',
    }

    def __init__(self, colors=None):
        if colors is not None:
            self.colors = self.colors.copy()
            self.colors.update(colors)

    def print(self, *args, **kwargs):
        """Like built-in ``print()`` but colorizes strings.

        Pass ``color`` as a keyword arg to colorize ``*args`` before
        printing them. If no ``color`` is passed, *args will printed
        without color.

        """
        color = kwargs.pop('color', None)
        file = kwargs.get('file', sys.stdout)
        if color and is_a_tty(file):
            string = self.string(color, *args, **kwargs)
            print(string, **kwargs)
        else:
            print(*args, **kwargs)

    def string(self, color, *args, **kwargs):
        """Returns a colorized string (joining ``args`` into one str).

        The arguments for this are similar to the built-in ``print()``.
        ``sep`` is a space by default, but ``end`` is an empty string.

        """
        color = self.colors[color]
        sep = kwargs.get('sep', ' ')
        end = kwargs.get('end', '')
        template = '{color}{string}{reset}'
        string = sep.join(str(a) for a in args)
        string = template.format(
            color=color, string=string, reset=self.colors['reset'])
        if end:
            string += end
        return string


color_printer = ColorPrinter()



import re
from collections import Mapping, Sequence
from itertools import chain

import six

from .util import NO_DEFAULT, NO_DEFAULT as PLACEHOLDER


class DottedAccessMixin:

    """Provides dotted access to nested items in dict-like containers.

    Mix with any type that supports item access to allow dotted access
    to nested items.

    >>> class MyMapping(dict, DottedAccessMixin):
    ...
    ...     pass
    >>>
    >>> my_map = MyMapping({'a': {'b': {'c': '123', '3': ['x', 'y', 'z']}}})

    Check whether an item is in the container, then get it:

    >>> my_map.contains_dotted('a.b.c')
    True
    >>> my_map.get_dotted('a.b.c')
    '123'

    Get an item from a sequence (the 3 is wrapped in parentheses to keep
    it from being interpreted as a sequence index):

    >>> my_map.contains_dotted('a.b.(3).0')
    True
    >>> my_map.get_dotted('a.b.(3).0')
    'x'

    Try getting an item that doesn't exist (without a default, this
    would raise a KeyError):

    >>> my_map.get_dotted('a.b.see', default=None)

    Add the missing item and then get it:

    >>> my_map.contains_dotted('a.b.see')
    False
    >>> my_map.set_dotted('a.b.see', 1)
    >>> my_map.contains_dotted('a.b.see')
    True
    >>> my_map.get_dotted('a.b.see')
    1

    """

    def contains_dotted(self, name):
        try:
            self._traverse(name)
        except KeyError:
            return False
        return True

    def get_dotted(self, name, default=NO_DEFAULT):
        try:
            return self._traverse(name)
        except KeyError:
            if default is NO_DEFAULT:
                raise
            return default

    def set_dotted(self, name, value, create_missing=True):
        self._traverse(name, create_missing=create_missing, value=value)

    def _traverse(self, name, create_missing=False, action=None, value=NO_DEFAULT):
        """Traverse to the item specified by ``name``.

        To create missing items on the way to the ``name``d item, pass
        ``create_missing=True``. This will insert an item for each
        missing segment in ``name``. The type and value of item that
        will be inserted for a missing segment depends on the *next*
        segment. If a ``default`` value is passed, the ``name``d item
        will be set to this value; otherwise, a default default will
        be used. See :meth:`_create_segment` for more info.

        """
        obj = self
        segments = self._parse_path(name)

        for segment, next_segment in zip(segments, segments[1:] + [None]):
            last = next_segment is None

            if create_missing:
                self._create_segment(obj, segment, next_segment)

            try:
                next_obj = obj[segment]
            except IndexError:
                raise KeyError(segment)

            if not last:
                obj = next_obj
            else:
                if action:
                    value = action(obj, segment)
                elif value is not NO_DEFAULT:
                    obj[segment] = value
                else:
                    value = obj[segment]

        return value

    def _create_segment(self, obj, segment, next_segment):
        """Create ``obj[segment]`` if missing.

        The default value for a missing segment is based on the *next*
        segment, unless a ``default`` is explicitly passed.

        If the next segment is an int, the default will be a list with
        the indicated number of items. Otherwise the default will be
        a :class:`Settings` dict.

        """
        if isinstance(next_segment, int):
            value = [PLACEHOLDER] * (next_segment + 1)
        else:
            value = Settings()
        if isinstance(obj, Mapping):
            if segment not in obj:
                obj[segment] = value
        elif isinstance(obj, Sequence):
            old_len = len(obj)
            new_len = segment + 1
            if new_len > old_len:
                obj.extend([PLACEHOLDER] * (new_len - old_len))
            if obj[segment] is PLACEHOLDER:
                obj[segment] = value

    def _parse_path(self, path):
        """Parse ``path`` into segments.

        Paths must start with a WORD (i.e., a top level Django setting
        name). Path segments are separated by dots. Compound path
        segments (i.e., a name with a dot in it) can be grouped inside
        parentheses.

        Examples::

            >>> settings = Settings()
            >>> settings._parse_path('WORD')
            ['WORD']
            >>> settings._parse_path('WORD.x')
            ['WORD', 'x']
            >>> settings._parse_path('WORD.(x)')
            ['WORD', 'x']
            >>> settings._parse_path('WORD.(x.y)')
            ['WORD', 'x.y']
            >>> settings._parse_path('WORD.(x.y).z')
            ['WORD', 'x.y', 'z']
            >>> settings._parse_path('WORD.0.z')
            ['WORD', 0, 'z']
            >>> settings._parse_path('WORD.(0).z')
            ['WORD', '0', 'z']
            >>> settings._parse_path('WORD.(0)X.z')
            ['WORD', '0X', 'z']

        An example of where compound names are actually useful is in
        logger settings::

            LOGGING.loggers.(package.module).handlers = ["console"]
            LOGGING.loggers.(package.module).level = "DEBUG"

        Paths may also contain interpolation groups. Dotted names in
        these groups will not be split (so there's no need to group them
        inside parentheses)::

            >>> settings = Settings()
            >>> settings._parse_path('WORD.{{x}}')
            ['WORD', '{{x}}']
            >>> settings._parse_path('WORD.{{x.y}}')
            ['WORD', '{{x.y}}']
            >>> settings._parse_path('WORD.{{x.y.z}}XYZ')
            ['WORD', '{{x.y.z}}XYZ']

        Interpolation groups *can* be wrapped in parentheses, but doing
        so is redundant::

            >>> settings._parse_path('WORD.({{x.y.z}}XYZ)')
            ['WORD', '{{x.y.z}}XYZ']

        Any segment that A) looks like an int and B) does *not* contain
        a (...) or {{...}} group will be converted to an int. Segments
        that start with a leading "0" followed by other digits will not
        be converted.

        """
        if not path:
            raise ValueError('path cannot be empty')

        segments = []
        path_iter = zip(iter(path), chain(path[1:], (None,)))
        if six.PY2:
            # zip() returns a list on Python 2
            path_iter = iter(path_iter)
        convert_name = self._convert_name
        current_segment = []
        current_segment_contains_group = False

        def append_segment():
            segment = ''.join(current_segment)
            if not current_segment_contains_group:
                segment = convert_name(segment)
            segments.append(segment)
            del current_segment[:]

        for c, d in path_iter:
            if c == '.':
                append_segment()
                current_segment_contains_group = False
            elif c == '(':
                nested = 0
                for c, d in path_iter:
                    current_segment.append(c)
                    if c == '(':
                        nested += 1
                    elif c == ')':
                        if nested:
                            nested -= 1
                        else:
                            current_segment.pop()  # Remove the closing paren
                            current_segment_contains_group = True
                            break
                else:
                    raise ValueError('Unclosed (...) in %s' % path)
            elif c == '{' and d == '{':
                current_segment_contains_group = True
                current_segment.append(c)
                for c, d in path_iter:
                    current_segment.append(c)
                    if c == '}' and d == '}':
                        current_segment_contains_group = True
                        break
                else:
                    raise ValueError('Unclosed {{...}} in %s' % path)
            else:
                current_segment.append(c)

        if current_segment:
            append_segment()

        return segments

    def _convert_name(self, name):
        """Convert ``name`` to int if it looks like an int.

        Otherwise, return it as is.

        """
        if re.search('^\d+$', name):
            if len(name) > 1 and name[0] == '0':
                # Don't treat strings beginning with "0" as ints
                return name
            return int(name)
        return name


class DottedAccessDict(dict, DottedAccessMixin):

    """Default implementation of a dict providing dotted item access.

    Typically used to wrap an existing dict to get dotted item access:

    >>> d = {'a': {'b': 'c'}}
    >>> d = DottedAccessDict(d)
    >>> d.get_dotted('a.b')
    'c'

    """


class Settings(dict, DottedAccessMixin):

    """Dict-ish container for settings.

    Provides access to settings via regular item access, attribute
    access, and dotted names::

        >>> settings = Settings()
        >>> settings['name'] = 'name'
        >>> settings['name']
        'name'
        >>> settings.name = 'name'
        >>> settings.name
        'name'
        >>> settings.set_dotted('NAMESPACE.name', 'nested name')
        >>> settings.get_dotted('NAMESPACE.name')
        'nested name'

    These are all equivalent::

        >>> settings['NAMESPACE']['name']
        'nested name'
        >>> settings.NAMESPACE.name
        'nested name'
        >>> settings.get_dotted('NAMESPACE.name')
        'nested name'

    Adding an item with a dotted name will create nested settings like
    so::

        >>> settings = Settings()
        >>> settings.set_dotted('NAMESPACE.name', 'nested name')
        >>> settings
        {'NAMESPACE': {'name': 'nested name'}}

    Implementation Notes
    ====================

    This is a subclass of dict because certain settings consumers, such
    as ``logging.dictConfig`` from the standard library, won't work with
    a non-dict mapping type because they do things like
    ``isinstance(something, dict)``.

    Where a setting has a name that's the same as an attribute name
    (e.g., ``get`` or ``update``), the attribute will take precedence.
    To get at such a setting, item access must be used. This is
    necessary because we can't allow settings to override the dict
    interface (because this might cause problems with outside consumers
    of the settings that aren't aware of the magic we're doing here).

    """

    def __init__(self, *args, **kwargs):
        # Call our update() instead of super().__init__() so that our
        # __setitem__() will be used.
        self.update(*args, **kwargs)

    def __setitem__(self, name, value):
        if isinstance(value, Mapping):
            value = Settings(value)
        super(Settings, self).__setitem__(name, value)

    # Implementation of attribute access.

    def __getattr__(self, name):
        # This is only invoked if the named attribute isn't found as an
        # instance or class attribute of the Settings instance. In other
        # words, this will only be called for settings stored in the
        # instance's internal dict storage. For methods such as `get`,
        # this won't be called.
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    # Support for copying and pickling.

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    # The following are required because these methods on the built-in
    # dict type will *not* call our __setitem__ method above. These
    # implementations were copied from collections.MutableMapping in the
    # standard library and tweaked slightly.

    def setdefault(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            self[name] = default
        return self[name]

    def update(*args, **kwargs):
        if len(args) > 2:
            raise TypeError(
                'update() takes at most 2 positional arguments ({} given)'.format(len(args)))
        elif not args:
            raise TypeError('update() takes at least 1 argument (0 given)')
        self = args[0]
        other = args[1] if len(args) >= 2 else ()
        if isinstance(other, Mapping):
            for name in other:
                self[name] = other[name]
        elif hasattr(other, 'keys'):
            for name in other.keys():
                self[name] = other[name]
        else:
            for name, value in other:
                self[name] = value
        for name, value in kwargs.items():
            self[name] = value



import io
import os


NO_DEFAULT = type('NO_DEFAULT', (), {
    '__nonzero__': (lambda self: False),  # Python 2
    '__bool__': (lambda self: False),  # Python 3
    '__str__': (lambda self: self.__class__.__name__),
    '__repr__': (lambda self: str(self)),
    '__copy__': (lambda self: self),
})()


def get_file_name():
    # We could set this as a global, but late binding is better in this
    # case. E.g., someone might import this module but not have set
    # os.environ['LOCAL_SETTINGS_FILE'] yet. We don't want them to have
    # to worry about the order of imports.
    file_name = os.environ.get('LOCAL_SETTINGS_FILE')
    if not file_name:
        file_name = os.path.join(os.getcwd(), 'local.cfg')
    return file_name


# These TTY functions were copied from Invoke


def is_a_tty(stream):
    if hasattr(stream, 'isatty') and callable(stream.isatty):
        return stream.isatty()
    elif has_fileno(stream):
        return os.isatty(stream.fileno())
    return False


def has_fileno(stream):
    try:
        return isinstance(stream.fileno(), int)
    except (AttributeError, io.UnsupportedOperation):
        return False



from collections import Mapping, MutableSequence, Sequence

from django.utils.module_loading import import_string

from six import string_types

from .base import Base
from .checker import Checker
from .settings import DottedAccessDict, Settings
from .types import LocalSetting

from .strategy import INIJSONStrategy


class Loader(Base):

    def __init__(self, file_name, section=None, registry=None, strategy_type=INIJSONStrategy):
        super(Loader, self).__init__(file_name, section, registry, strategy_type)

    def load_and_check(self, base_settings, prompt=None):
        """Load settings and check them.

        Loads the settings from ``base_settings``, then checks them.

        Returns:
            (merged settings, True) on success
            (None, False) on failure

        """
        checker = Checker(self.file_name, self.section, self.registry, self.strategy_type, prompt)
        settings = self.load(base_settings)
        if checker.check(settings):
            return settings, True
        return None, False

    def load(self, base_settings):
        """Merge local settings from file with ``base_settings``.

        Returns a new settings dict containing the base settings and the
        loaded settings. Includes:

            - base settings
            - settings from extended file(s), if any
            - settings from file

        """
        is_valid_key = lambda k: k.isupper() and not k.startswith('_')

        # Base settings, including `LocalSetting`s, loaded from the
        # Django settings module.
        valid_keys = (k for k in base_settings if is_valid_key(k))
        base_settings = DottedAccessDict((k, base_settings[k]) for k in valid_keys)

        # Settings read from the settings file; values are unprocessed.
        settings_from_file = self.strategy.read_file(self.file_name, self.section)
        settings_from_file.pop('extends', None)

        # The fully resolved settings.
        settings = Settings(base_settings)

        settings_names = []
        settings_not_decoded = set()

        for name, value in settings_from_file.items():
            for prefix in ('PREPEND.', 'APPEND.', 'SWAP.'):
                if name.startswith(prefix):
                    name = name[len(prefix):]
                    name = '{prefix}({name})'.format(**locals())
                    break

            settings_names.append(name)

            # Attempt to decode raw values. Errors in decoding at this
            # stage are ignored.
            try:
                value = self.strategy.decode_value(value)
            except ValueError:
                settings_not_decoded.add(name)

            settings.set_dotted(name, value)

            # See if this setting corresponds to a `LocalSetting`. If
            # so, note that the `LocalSetting` has a value by putting it
            # in the registry. This also makes it easy to retrieve the
            # `LocalSetting` later so its value can be set.
            current_value = base_settings.get_dotted(name, None)
            if isinstance(current_value, LocalSetting):
                self.registry[current_value] = name

        # Interpolate values of settings read from file. When a setting
        # that couldn't be decoded previously is encountered, its post-
        # interpolation value will be decoded.
        for name in settings_names:
            value = settings.get_dotted(name)
            value, _ = self._interpolate_values(value, settings)
            if name in settings_not_decoded:
                value = self.strategy.decode_value(value)
            settings.set_dotted(name, value)

        # Interpolate base settings.
        self._interpolate_values(settings, settings)

        self._interpolate_keys(settings, settings)

        self._prepend_extras(settings, settings.pop('PREPEND', None))
        self._append_extras(settings, settings.pop('APPEND', None))
        self._swap_list_items(settings, settings.pop('SWAP', None))
        self._import_from_string(settings, settings.pop('IMPORT_FROM_STRING', None))

        for local_setting, name in self.registry.items():
            local_setting.value = settings.get_dotted(name)

        return settings

    # Post-processing

    def _interpolate_values(self, obj, settings):
        all_interpolated = []
        interpolated = []
        while interpolated is not None:
            all_interpolated.extend(interpolated)
            obj, interpolated = self._interpolate_values_inner(obj, settings)
        return obj, all_interpolated

    def _interpolate_values_inner(self, obj, settings, _interpolated=None):
        if _interpolated is None:
            _interpolated = []

        if isinstance(obj, string_types):
            new_value, changed = self._inject(obj, settings)
            if changed:
                _interpolated.append((obj, new_value))
                obj = new_value
        elif isinstance(obj, Mapping):
            for k, v in obj.items():
                obj[k], _interpolated = self._interpolate_values_inner(v, settings, _interpolated)
        elif isinstance(obj, MutableSequence):
            for i, v in enumerate(obj):
                obj[i], _interpolated = self._interpolate_values_inner(v, settings, _interpolated)
        elif isinstance(obj, Sequence):
            items = []
            for v in obj:
                item, _interpolated = self._interpolate_values_inner(v, settings, _interpolated)
                items.append(item)
            obj = obj.__class__(items)

        return obj, _interpolated or None

    def _interpolate_keys(self, obj, settings):
        if isinstance(obj, Mapping):
            replacements = {}
            for k, v in obj.items():
                if isinstance(k, string_types):
                    new_k, changed = self._inject(k, settings)
                    if changed:
                        replacements[k] = new_k
                self._interpolate_keys(v, settings)
            for k, new_k in replacements.items():
                obj[new_k] = obj[k]
                del obj[k]
        elif isinstance(obj, Sequence) and not isinstance(obj, string_types):
            for item in obj:
                self._interpolate_keys(item, settings)

    def _prepend_extras(self, settings, extras):
        if not extras:
            return
        for name, extra_val in extras.items():
            if not extra_val:
                continue
            current_val = settings.get_dotted(name)
            if not isinstance(current_val, Sequence):
                raise TypeError('PREPEND only works with list-type settings')
            settings.set_dotted(name, extra_val + current_val)

    def _append_extras(self, settings, extras):
        if not extras:
            return
        for name, extra_val in extras.items():
            if not extra_val:
                continue
            current_val = settings.get_dotted(name)
            if not isinstance(current_val, Sequence):
                raise TypeError('APPEND only works with list-type settings')
            settings.set_dotted(name, current_val + extra_val)

    def _swap_list_items(self, settings, swap):
        if not swap:
            return
        for name, swap_map in swap.items():
            if not swap_map:
                continue
            current_val = settings.get_dotted(name)
            if not isinstance(current_val, Sequence):
                raise TypeError('SWAP only works with list-type settings')
            for old_item, new_item in swap_map.items():
                k = current_val.index(old_item)
                current_val[k] = new_item

    def _import_from_string(self, settings, import_from_string):
        if not import_from_string:
            return
        for name in import_from_string:
            current_val = settings.get_dotted(name)
            if isinstance(current_val, string_types):
                settings.set_dotted(name, import_string(current_val))

    def _inject(self, value, settings):
        """Inject ``settings`` into ``value``.

        Go through ``value`` looking for ``{{NAME}}`` groups and replace
        each group with the value of the named item from ``settings``.

        Args:
            value (str): The value to inject settings into
            settings: An object that provides the dotted access interface

        Returns:
            (str, bool): The new value and whether the new value is
                different from the original value

        """
        assert isinstance(value, string_types), 'Expected str; got {0.__class__}'.format(value)

        begin, end = '{{', '}}'

        if begin not in value:
            return value, False

        new_value = value
        begin_pos, end_pos = 0, None
        len_begin, len_end = len(begin), len(end)
        len_value = len(new_value)

        while begin_pos < len_value:
            # Find next {{.
            begin_pos = new_value.find(begin, begin_pos)

            if begin_pos == -1:
                break

            # Save everything before {{.
            before = new_value[:begin_pos]

            # Find }} after {{.
            begin_pos += len_begin
            end_pos = new_value.find(end, begin_pos)
            if end_pos == -1:
                raise ValueError('Unmatched {begin}...{end} in {value}'.format(**locals()))

            # Get name between {{ and }}, ignoring leading and trailing
            # whitespace.
            name = new_value[begin_pos:end_pos]
            name = name.strip()

            if not name:
                raise ValueError('Empty name in {value}'.format(**locals()))

            # Save everything after }}.
            after_pos = end_pos + len_end
            try:
                after = new_value[after_pos:]
            except IndexError:
                # Reached end of value.
                after = ''

            # Retrieve string value for named setting (the "injection
            # value").
            try:
                injection_value = settings.get_dotted(name)
            except KeyError:
                raise KeyError('{name} not found in {settings}'.format(**locals()))

            if not isinstance(injection_value, string_types):
                injection_value = self.strategy.encode_value(injection_value)

            # Combine before, inject value, and after to get the new
            # value.
            new_value = ''.join((before, injection_value, after))

            # Continue after injected value.
            begin_pos = len(before) + len(injection_value)
            len_value = len(new_value)

        return new_value, (new_value != value)



class LocalSettingsError(Exception):

    pass


class SettingsFileNotFoundError(LocalSettingsError):

    pass


class SettingsFileSectionNotFoundError(LocalSettingsError, LookupError):

    pass


class SettingsFileDidNotPassCheck(LocalSettingsError):

    pass


class NoDefaultError(LocalSettingsError):

    pass


class NoValueError(LocalSettingsError):

    pass



import sys
from collections import Mapping, Sequence

from six import string_types
from six.moves import input

from .base import Base
from .color_printer import color_printer as printer
from .strategy import INIJSONStrategy
from .types import LocalSetting
from .util import NO_DEFAULT, is_a_tty


class Checker(Base):

    def __init__(self, file_name, section=None, registry=None, strategy_type=INIJSONStrategy,
                 prompt=None):
        super(Checker, self).__init__(file_name, section, registry, strategy_type)
        if prompt is None:
            try:
                prompt = is_a_tty(sys.stdin) and is_a_tty(sys.stdout)
            except AttributeError:
                prompt = False
        self.prompt = prompt

    def check(self, obj, prefix=None):
        """Recursively look for :class:`.LocalSetting`s in ``obj``.

        ``obj`` can be a dict, tuple, or list. Other types are skipped.

        This will prompt to get the value of local settings (excluding
        those that have had prompting disabled) that haven't already
        been set locally.

        Returns ``True`` or ``False`` to indicate whether settings were
        successfully checked.

        """
        self._populate_registry(obj, prefix)
        settings_to_write, missing = self._check(obj, prefix, {}, {})
        if settings_to_write:
            self.strategy.write_settings(settings_to_write, self.file_name, self.section)
        if missing:
            for name, local_setting in missing.items():
                printer.print_error('Local setting `{name}` must be set'.format(name=name))
            return False
        return True

    def _populate_registry(self, obj, prefix=None):
        if isinstance(obj, Mapping):
            items = obj.items()
        elif isinstance(obj, Sequence) and not isinstance(obj, string_types):
            items = zip(range(len(obj)), obj)
        else:
            return
        for k, v in items:
            name = k if prefix is None else '{0}.{1}'.format(prefix, k)
            if not isinstance(v, LocalSetting):
                self._populate_registry(v, name)
            else:
                self.registry[v] = name

    def _check(self, obj, prefix, settings_to_write, missing):
        if isinstance(obj, Mapping):
            items = sorted(obj.items(), key=lambda item: item[0])
        elif isinstance(obj, Sequence) and not isinstance(obj, string_types):
            items = zip(range(len(obj)), obj)
        else:
            return {}, {}

        for k, v in items:
            name = k if not prefix else '{0}.{1}'.format(prefix, k)
            if not isinstance(v, LocalSetting):
                self._check(v, name, settings_to_write, missing)
            elif k not in settings_to_write:
                # Note: If k is already in settings_to_write, this local
                # setting was set as a result of being another setting's
                # default.
                is_set = False
                local_setting = v

                if local_setting.derived_default:
                    # Ensure this setting's default is set if it's also a local setting
                    default_name = self.registry[local_setting.derived_default]
                    if not local_setting.has_default:
                        self._check(
                            {default_name: local_setting.derived_default},
                            None, settings_to_write, missing)

                if self.prompt:
                    printer.print_header('=' * 79)

                if local_setting.prompt:  # prompt for value
                    if self.prompt:
                        v, is_set = self.prompt_for_value(name, v)
                elif local_setting.has_default:  # use default w/o prompting
                    v, is_set = local_setting.default, True
                    msg = (
                        'Using default value `{value!r}` for local setting '
                        '`{name}`'.format(name=name, value=v))
                    if local_setting.derived_default:
                        msg += ' (derived from {0})'.format(default_name)
                    printer.print_warning(msg)

                if is_set:
                    local_setting.value = obj[k] = settings_to_write[name] = v
                else:
                    missing[name] = v

        return settings_to_write, missing

    def prompt_for_value(self, name, local_setting):
        v, is_set = NO_DEFAULT, False
        while not is_set:  # Keep prompting until valid value is set
            printer.print_header(
                'Enter a value for the local setting `{name}` (as JSON)'.format(name=name))
            if local_setting.doc:
                printer.print_header(local_setting.doc)
            if local_setting.has_default:
                msg = 'Hit enter to use default: `{0!r}`'.format(local_setting.default)
                if local_setting.derived_default:
                    default_name = self.registry[local_setting.derived_default]
                    msg += ' (derived from {0})'.format(default_name)
                printer.print_header(msg)
            v = input('> ').strip()
            if v:
                try:
                    v = self.strategy.decode_value(v)
                except ValueError as e:
                    printer.print_error(e)
                else:
                    is_set = local_setting.validate(v)
                    if not is_set:
                        printer.print_error('`{0}` is not a valid value for {1}'.format(v, name))
            elif local_setting.has_default:
                v, is_set = local_setting.default, True
                printer.print_info('Using default value for `{0}`'.format(name))
            else:
                printer.print_error('You must enter a value for `{0}`'.format(name))
        return v, is_set



import json

from .exc import NoDefaultError, NoValueError
from .util import NO_DEFAULT


class LocalSetting(object):

    """Used to mark settings that should be set in the local environment.

    If a default value isn't specified, the local setting must be set.
    Generally, local settings with no default value should not be stored
    in version control.

    If a local setting doesn't have a default value, it will be prompted
    for by default. Pass ``prompt=False`` to disable this (or use
    :class:`.SecretSetting`).

    """

    def __init__(self, default=NO_DEFAULT, prompt=True, doc=None, validator=None):
        self.default = default
        self.derived_default = NO_DEFAULT
        if default is not NO_DEFAULT:
            if isinstance(default, LocalSetting):
                self.derived_default = default
            else:
                if callable(default):
                    default = default()
                try:
                    json.dumps(default)
                except TypeError as exc:
                    msg = (
                        '{exc}\nDefault value for LocalSetting must be JSON '
                        'serializable'.format(exc=exc))
                    raise TypeError(msg)
        self.prompt = prompt
        self.doc = doc
        self.validator = validator
        self._value = NO_DEFAULT

    @property
    def has_default(self):
        return self._get_default() is not NO_DEFAULT

    @property
    def has_value(self):
        return self._value is not NO_DEFAULT or self.has_default

    def _get_default(self):
        if self.derived_default:
            default = self.derived_default.value
        else:
            default = self._default
            if callable(default):
                default = default()
        return default

    @property
    def default(self):
        default = self._get_default()
        if default is NO_DEFAULT:
            raise NoDefaultError('Local setting has no default value')
        return default

    @default.setter
    def default(self, default):
        self._default = default

    @property
    def value(self):
        value = self._value
        if value is NO_DEFAULT:
            value = self._get_default()
        if value is NO_DEFAULT:
            raise NoValueError('Local setting has no value')
        return value

    @value.setter
    def value(self, value):
        if not self.validate(value):
            raise ValueError('`{0}` is not a valid value'.format(value))
        self._value = value

    def validate(self, v):
        if self.validator:
            return self.validator(v)
        return True

    def __str__(self):
        try:
            default = repr(self.default)
        except NoDefaultError:
            default = '[NO DEFAULT VALUE]'
        try:
            value = repr(self.value)
        except NoValueError:
            value = '[NO VALUE SET]'
        s = '<{0.__class__.__name__} with default `{default}` and value `{value}`>'
        s = s.format(self, default=default, value=value)
        return s


class SecretSetting(LocalSetting):

    """Used to mark secret settings.

    Secret settings have no default and should never be stored in
    version control. They will always be prompted for if not present in
    the local settings file.

    """

    def __init__(self, doc=None, validator=None):
        super(SecretSetting, self).__init__(NO_DEFAULT, True, doc, validator)



class Base(object):

    def __init__(self, file_name, section=None, registry=None, strategy_type=None):
        strategy = strategy_type()
        file_name, section = strategy.parse_file_name_and_section(file_name, section)
        self.file_name = file_name
        self.section = section
        # Registry of local settings with a value in the settings file
        self.registry = {} if registry is None else registry
        self.strategy_type = strategy_type
        self.strategy = strategy



import argparse
import importlib
import os
import sys
import textwrap

from setuptools import find_packages

from .color_printer import color_printer as printer
from .loader import Loader
from .strategy import get_file_type_map, guess_strategy_type


class ArgParser(argparse.ArgumentParser):

    def error(self, message):
        message = '{self.prog} error: {message}\n'.format(**locals())
        message = printer.string_error(message)
        self.print_usage(sys.stderr)
        self.exit(2, message)


def make_local_settings(argv=None):
    """Generate a local settings file.

    The most common usage is:

        make-local-settings <env>

    where env is replaced with and environment name such as stage or
    prod. For example:

        make-local-settings prod

    This will create a local settings file named local.prod.cfg with
    a prod section if local.prod.cfg does not exist. If local.prod.cfg
    already exists, the prod section will be added if it doesn't exist,
    and any missing local settings will be prompted for and added.

    """
    description = textwrap.dedent('    %s' % make_local_settings.__doc__)

    parser = ArgParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument('env', nargs='?', help='Environment name (e.g., dev or prod)')
    parser.add_argument('-t', '--type', default='cfg')
    parser.add_argument(
        '-b', '--base-settings-module', default=None,
        help='Base settings module as a dotted path')
    parser.add_argument(
        '-f', '--file-name', default=None,
        help='Name of file to write settings into')
    parser.add_argument(
        '-s', '--section', default=None,
        help='Section to read/write settings from/to (default = <env>)')
    parser.add_argument(
        '-e', '--extends', default=None,
        help='File name and section to extend from as file_name.ext#section (section is optional)')
    parser.add_argument('-o', '--overwrite', action='store_true', default=False)

    args = parser.parse_args(argv)

    if args.type:
        strategy_type = get_file_type_map()[args.type]
        strategy = strategy_type()

    if args.base_settings_module is None:
        package = find_packages()[0]
        path = os.path.join(os.getcwd(), package, 'settings.py')
        if os.path.exists(path):
            args.base_settings_module = '{package}.settings'.format(package=package)
            printer.print_info(
                'Using {0.base_settings_module} as base settings module'.format(args))
        else:
            parser.error('Could not guess which base settings module to use; specify with -b')

    if args.file_name:
        file_name = args.file_name
        if not args.type:
            strategy_type = guess_strategy_type(file_name)
            strategy = strategy_type()
    elif args.env:
        file_name = 'local.{0.env}.{ext}'.format(args, ext=strategy.file_types[0])
    else:
        parser.error('Either env or file name must be specified')
    file_name = os.path.normpath(os.path.abspath(file_name))

    if args.section:
        section = args.section
    elif args.env:
        section = args.env
    else:
        parser.error('Either env or section must be specified')

    if os.path.exists(file_name):
        if args.overwrite:
            printer.print_warning('Overwriting', file_name)
            os.remove(file_name)

    # This will create the file if it doesn't exist. The extends
    # settings will be written whether or not the file exists.
    settings = {}
    if args.extends:
        settings['extends'] = args.extends
    strategy.write_settings(settings, file_name, section)

    # Load base settings from settings module while A) ensuring that
    # local settings aren't loaded in the process and B) accounting for
    # the fact that the base settings might not load local settings.
    original_disable_value = os.environ.get('LOCAL_SETTINGS_CONFIG_DISABLE')
    os.environ['LOCAL_SETTINGS_CONFIG_DISABLE'] = '1'
    settings_module = importlib.import_module(args.base_settings_module)
    base_settings = vars(settings_module)
    os.environ.pop('LOCAL_SETTINGS_CONFIG_DISABLE')
    if original_disable_value is not None:
        os.environ['LOCAL_SETTINGS_CONFIG_DISABLE'] = original_disable_value

    loader = Loader(file_name, section, strategy_type=strategy_type)
    loader.load_and_check(base_settings, prompt=True)


if __name__ == '__main__':
    make_local_settings()



import sys

from setuptools import find_packages, setup


py_version = sys.version_info[:2]
py_version_dotted = '{0.major}.{0.minor}'.format(sys.version_info)
supported_py_versions = ('2.7', '3.3', '3.4', '3.5', '3.6')


if py_version_dotted not in supported_py_versions:
    sys.stderr.write('WARNING: django-local-settings does not officially support Python ')
    sys.stderr.write(py_version_dotted)
    sys.stderr.write('\n')


with open('VERSION') as version_fp:
    VERSION = version_fp.read().strip()


with open('README.md') as readme_fp:
    long_description = readme_fp.read()


install_requires = [
    'six',
]

if py_version < (3, 0):
    install_requires.append('configparser')

# NOTE: Keep this Django version up to date with the latest Django
#       release that works for the versions of Python we support.
#       This is used to get up and running quickly; tox is used to test
#       all supported Python/Django version combos.
if py_version == (3, 3):
    django_spec = 'django>=1.8,<1.9',
else:
    django_spec = 'django>=1.10,<1.11',


setup(
    name='django-local-settings',
    version=VERSION,
    author='Wyatt Baldwin',
    author_email='wbaldwin@pdx.edu',
    url='https://github.com/PSU-OIT-ARC/django-local-settings',
    description='A system for dealing with local settings in Django projects',
    long_description=long_description,
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
        'dev': [
            'coverage>=4',
            django_spec,
            'flake8',
            'tox>=2.6.0',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ] + [
        'Programming Language :: Python :: {v}'.format(v=v)for v in supported_py_versions
    ],
    entry_points="""
    [console_scripts]
    make-local-settings = local_settings:make_local_settings

    """,
)



import doctest
import unittest

from .. import settings
from ..settings import Settings


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(settings))
    return tests


class TestPathParsing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.settings = Settings()

    def test_simple_path(self):
        segments = self.settings._parse_path('XYZ')
        self.assertEqual(segments, ['XYZ'])

    def test_dotted_path(self):
        segments = self.settings._parse_path('XYZ.abc')
        self.assertEqual(segments, ['XYZ', 'abc'])

    def test_multi_dotted_path(self):
        segments = self.settings._parse_path('XYZ.abc.x.y.z')
        self.assertEqual(segments, ['XYZ', 'abc', 'x', 'y', 'z'])

    def test_compound_path_at_end(self):
        segments = self.settings._parse_path('XYZ.(a.b.c)')
        self.assertEqual(segments, ['XYZ', 'a.b.c'])

    def test_compound_path_in_middle(self):
        segments = self.settings._parse_path('XYZ.(a.b.c).d')
        self.assertEqual(segments, ['XYZ', 'a.b.c', 'd'])

    def test_non_dotted_compound_path(self):
        segments = self.settings._parse_path('XYZ.(abc)')
        self.assertEqual(segments, ['XYZ', 'abc'])

    def test_multi_non_dotted_compound_path_at_end(self):
        segments = self.settings._parse_path('XYZ.(a).(b).(c)')
        self.assertEqual(segments, ['XYZ', 'a', 'b', 'c'])

    def test_multi_non_dotted_compound_path_in_middle(self):
        segments = self.settings._parse_path('XYZ.(a).(b).(c).dddd')
        self.assertEqual(segments, ['XYZ', 'a', 'b', 'c', 'dddd'])

    def test_complex_path(self):
        segments = self.settings._parse_path('XYZ.(a).(b.b).c.(d)')
        self.assertEqual(segments, ['XYZ', 'a', 'b.b', 'c', 'd'])


class TestSetItem(unittest.TestCase):

    def setUp(self):
        self.settings = Settings()

    def test_setitem(self):
        self.settings['a'] = 1
        self.assertIn('a', self.settings)
        self.assertEqual(self.settings['a'], 1)

    def test_set_dotted(self):
        self.settings.set_dotted('a.b', 1)
        self.assertIn('a', self.settings)
        self.assertIn('b', self.settings['a'])
        self.assertEqual(self.settings['a']['b'], 1)
        self.assertEqual(self.settings.a.b, 1)

    def test_setattr(self):
        self.settings.a = Settings()
        self.settings.a.b = 1
        self.assertIn('a', self.settings)
        self.assertIn('b', self.settings['a'])
        self.assertEqual(self.settings['a']['b'], 1)

    def test_setdefault(self):
        x = self.settings.setdefault('x', 1)
        self.assertIn('x', self.settings)
        self.assertEqual(x, 1)


class TestGetItem(unittest.TestCase):

    def setUp(self):
        self.settings = Settings({'a': {'b': {'c': 'c'}}})

    def test_getitem(self):
        self.assertEqual(self.settings['a']['b']['c'], 'c')

    def test_getitem_via_dotted_name(self):
        self.assertEqual(self.settings.get_dotted('a.b.c'), 'c')

    def test_getattr(self):
        self.assertEqual(self.settings.a.b.c, 'c')

    def test_get_missing(self):
        self.assertRaises(KeyError, self.settings.__getitem__, ('xxx',))

    def test_get_missing_dotted(self):
        self.assertRaises(KeyError, self.settings.__getitem__, ('x.y.z',))


class TestDotted(unittest.TestCase):

    def setUp(self):
        self.settings = Settings()

    def test_dotted(self):
        self.settings.set_dotted('x.y.z', 'z')
        self.assertTrue(self.settings.contains_dotted('x.y.z'))
        self.assertEqual(self.settings.get_dotted('x.y.z'), 'z')
        self.assertIn('x', self.settings)
        self.assertIn('y', self.settings['x'])
        self.assertIn('z', self.settings['x']['y'])
        self.assertEqual(self.settings['x']['y']['z'], 'z')
        self.assertEqual(self.settings.x.y.z, 'z')

    def test_nested_parentheses(self):
        self.settings.set_dotted('(x(y))', 'x')
        self.assertTrue(self.settings.contains_dotted('(x(y))'))
        self.assertEqual(self.settings.get_dotted('(x(y))'), 'x')
        self.assertIn('x(y)', self.settings)
        self.assertEqual(self.settings['x(y)'], 'x')

    def test_nested_parentheses_with_dots(self):
        self.settings.set_dotted('(x.(y.z))', 'x')
        self.assertTrue(self.settings.contains_dotted('(x.(y.z))'))
        self.assertEqual(self.settings.get_dotted('(x.(y.z))'), 'x')
        self.assertIn('x.(y.z)', self.settings)
        self.assertEqual(self.settings['x.(y.z)'], 'x')

    def test_nested_empty_parentheses(self):
        self.settings.set_dotted('(())', 'x')
        self.assertTrue(self.settings.contains_dotted('(())'))
        self.assertEqual(self.settings.get_dotted('(())'), 'x')
        self.assertIn('()', self.settings)
        self.assertEqual(self.settings['()'], 'x')

    def test_brackets(self):
        self.settings.set_dotted('{{x}}.y', 'x')
        self.assertTrue(self.settings.contains_dotted('{{x}}.y'))
        self.assertEqual(self.settings.get_dotted('{{x}}.y'), 'x')
        self.assertIn('{{x}}', self.settings)
        self.assertIn('y', self.settings['{{x}}'])
        self.assertEqual(self.settings['{{x}}']['y'], 'x')

    def test_brackets_with_internal_dots(self):
        self.settings.set_dotted('{{x.y.z}}', 'x')
        self.assertTrue(self.settings.contains_dotted('{{x.y.z}}'))
        self.assertEqual(self.settings.get_dotted('{{x.y.z}}'), 'x')
        self.assertIn('{{x.y.z}}', self.settings)
        self.assertEqual(self.settings['{{x.y.z}}'], 'x')

    def test_empty_brackets(self):
        self.settings.set_dotted('{{}}', 'x')
        self.assertTrue(self.settings.contains_dotted('{{}}'))
        self.assertEqual(self.settings.get_dotted('{{}}'), 'x')
        self.assertIn('{{}}', self.settings)
        self.assertEqual(self.settings['{{}}'], 'x')



import os
import unittest

from ..checker import Checker
from ..loader import Loader
from ..types import LocalSetting


LOCAL_SETTINGS_FILE = os.path.join(os.path.dirname(__file__), 'local.cfg#test')


class TestChecker(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.checker = Checker(LOCAL_SETTINGS_FILE)
        cls.loader = Loader(LOCAL_SETTINGS_FILE)

    def test_checker(self):
        local_setting = LocalSetting('default value')
        settings = self.loader.load({
            'LOCAL_SETTING': local_setting,
        })
        self.checker.check(settings)



import unittest

from ..types import SecretSetting


class TestSecretSetting(unittest.TestCase):

    def test_create_secret_setting(self):
        setting = SecretSetting()
        self.assertFalse(setting.has_default)
        self.assertFalse(setting.has_value)

    def test_secret_setting_cannot_have_a_default(self):
        self.assertRaises(TypeError, SecretSetting, default='pants')
